import threading
import tempfile
import os
import zipfile
import logging

from string import Template

from django.core.mail import send_mail
from standalone_cellpublisher import wrapper
from web_interface.create_diagram.models import UserSubmission, FinishedJob
from web_interface import settings


logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    filename=settings.JOBTHREADS_LOG_FILE,
                    level=logging.DEBUG)

class Job(threading.Thread):
    """
    Delegates the generation of the diagrams to the wrapper program based
    on a submission entry in the database.

    Each calculation runs on an individual thread, allowing the caller
    function to keep on running without the need to wait for the result of
    the calculation.

    Due to the fact that calculations are costly in ressources, only one
    thread at a time is allowed to access the wrapper method.

    """

    wrapperLock = threading.Lock()

    def __init__(self, submission_id):
        threading.Thread.__init__(self)
        self.submission_id = submission_id
        self.parameters = dict()
        
    def run(self):
        """
        Create two versions of the output of the standalone cellpublisher
        in the servers folders. The first version is a downloadable zip file
        and the second a preview of how the final Pathway looks like.

        """
        self._parse_database_entry()
        Job.wrapperLock.acquire()
        try:
            logging.debug("Running the submitted job...")
            result = wrapper.execute(self.parameters)
            logging.debug("Done!")

            logging.debug("Creating the zip file...")
            pathToZipFile = os.path.join(settings.ZIP_DOWNLOAD_ROOT,
                                         str(self.submission.url_id) + ".zip")
            self.zip_dir(self.targetFolder, pathToZipFile)            
            logging.debug("Done!")

            logging.debug("Creating the preview...")
            # Overwrite the markers.xml file including the PDB files
            wrapper.create_xml(result['offset'],
                               self.parameters['targetFolder'],
                               self.parameters['pathToSBML'],
                               True  # Allow download of PDBs
                               )

            logging.debug("Done!")

            logging.debug("Saving to the database...")
            self._register_result(ok=True, maxZoom=result['maxZoom'])
            logging.debug("Done!")
            
        except:
            logging.exception(
                "Could not process submission %s" % self.submission_id)
            self._register_result(ok=False)
            raise
        finally:
            Job.wrapperLock.release()
            
    def _parse_database_entry(self):
        """
        Handle the conversion between the database entry for the submission
        and the settings needed by the wrapper module.

        """
        try:
            self.submission = UserSubmission.objects.get(
                user_admin_id=self.submission_id)
            submission = self.submission
            self.parameters['apiKey'] = submission.maps_api_key
            self.targetFolder = os.path.join(settings.PREVIEW_ROOT,
                                        str(submission.url_id))
            os.mkdir(self.targetFolder)
            self.parameters['targetFolder'] = self.targetFolder
            self.parameters['pathToImage'] = submission.png.path
            self.parameters['pathToSBML'] = submission.sbml.path
            self.parameters['title'] = submission.html_title
            self.parameters['author'] = submission.owner
            self.parameters['email'] = submission.email
            self.parameters['downloadPDBs'] = False

            logging.debug(self.parameters)
            
        except:
            msg = "Could not parse submission %s " % self.submission_id
            msg += "from the database."
            logging.exception(msg) 
            self._register_result(ok=False)
            raise

        # More code to parse the submission to the right format for wrapper.py

    def zip_dir(self, dirpath, zippath):
        """
        Create a zip file with the contents in the 'dirpath'. Code written
        by Dmitry Nedbaylo.

        """
        fzip = zipfile.ZipFile(zippath, 'w', zipfile.ZIP_DEFLATED)
        basedir = os.path.join(os.path.dirname(dirpath), "") 
        for root, dirs, files in os.walk(dirpath):
            dirname = root.replace(basedir, '')
            for f in files:
                fzip.write(os.path.join(root, f),
                           os.path.join(dirname, f))
        fzip.close()

    def _register_result(self, ok, maxZoom=0):
        """
        Handle saving the finished job into the database.

        """
        db_entry = FinishedJob(submission = self.submission,
                               ok = ok,
                               maxZoom = maxZoom)
        db_entry.save()

        if ok:
            self._send_confirmation_email()

    def _send_confirmation_email(self):

        if not self.parameters['email']:
            return
        
        subject = "Conversion finished"

        template = """
Dear $author,

The CellDesigner diagram titled "$title" has been successfully converted to CellPublisher.

To go to the interactive diagram visit this link:
http://cellpublisher.gobics.de$url_preview

You can also visit the admin site for your diagram, where you can download and delete the pathway:
http://cellpublisher.gobics.de$url_admin

Kind regards,

The CellPublisher team
"""

        fields = { 'author':self.parameters['author'],
                   'title': self.parameters['title'],
                   'url_preview': self.submission.get_absolute_url(),
                   'url_admin': self.submission.get_admin_url()
                }
        

        send_mail(subject, Template(template).substitute(fields),
                  'CellPublisher <lflorez@gwdg.de>',
                  [self.parameters['email']], fail_silently=False)

    #def start(self):
     #   self.run()
