#!/usr/bin/python

import os
import shutil

from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import render_to_response
from django.http import Http404
from django.conf import settings

from models import UserSubmission, UserSubmissionForm, FinishedJob

from jobthreads import Job

def viewPathway(request, url_id):
    """
    Show the interactive pathway to the user.

    """
    submission = UserSubmission.objects.get(url_id=url_id)
    job = FinishedJob.objects.get(submission=submission)
    if job.deleted or not job.ok:
        raise Http404
    
    maxZoom = job.maxZoom
    return render_to_response('pathway.html',
                              {'submission': submission,
                               'maxZoom': maxZoom}
                              )

def discuss(request, url_id):
    """
    Show a discussion page for a diagram where users can post comments

    """
    try:
        submission = UserSubmission.objects.get(url_id=url_id)
    except UserSubmission.DoesNotExist:
        raise Http404
    title = submission.html_title
    return render_to_response("comments.html", {'submission':submission} )

def delete(request, user_admin_id):

    # Test if the URL belongs to a previous submission
    try:
        job = FinishedJob.objects.get(
                submission__user_admin_id=user_admin_id)
    except FinishedJob.DoesNotExist:
        raise Http404

    if job.deleted:
        pass
    else:
        # Delete the input files from the user
        try:
            job.submission.sbml.delete()
            job.submission.png.delete()
        except:
            pass
        
        # Delete the pathway on the server
        output_path_preview = os.path.join(settings.PREVIEW_ROOT,
                                           job.submission.url_id)
        if os.path.isdir(output_path_preview):
            shutil.rmtree(output_path_preview)
        
        # Delete the downloadable zip file
        output_path_zip = os.path.join(settings.ZIP_DOWNLOAD_ROOT,
                                       job.submission.url_id + ".zip")
        if os.path.isfile(output_path_zip):
            os.remove(output_path_zip)
                
        # Mark the pathway as deleted
        job.deleted = True
        job.save()
    return HttpResponseRedirect('/status/%s' % user_admin_id)

def hello(request):
    return HttpResponse("Hello world!")

def submit(request):
    """
    A view to handle the submission of jobs to the cellpublisher web interface.

    It displays a form based on the UserSubmission model, validates it and
    then saves it in the database. If the file was succesfully saved, the
    user is redirected to a status page for the submission.

    """

    # If the form was already submitted...
    if request.method == 'POST':
        # Add the IP-Address of the submitter to the model instance
        submission = UserSubmission(ip=request.META.get('REMOTE_ADDR',
                                                        '0.0.0.0'))
        # Load the form
        form = UserSubmissionForm(request.POST,
                                  request.FILES,
                                  instance=submission)
        if form.is_valid():
            user_admin_id = submission.user_admin_id
            form.save()
            myJob = Job(user_admin_id)
            myJob.start()
            return HttpResponseRedirect('/status/%s' % user_admin_id)

    # If the form was not yet submitted, show it to the user.
    else:
        form = UserSubmissionForm()

    return render_to_response('submission.html', {'form':form}) 

def status(request, submission_id):
    """A view that displays the status of a submitted job."""
    
    # Test if the URL belongs to an existing submission...
    try:
        submission = UserSubmission.objects.get(user_admin_id=submission_id)
    except UserSubmission.DoesNotExist:
        raise Http404

    # Test if the submission has been processed completely...
    try:
        job = FinishedJob.objects.get(submission__user_admin_id=submission_id)
    except FinishedJob.DoesNotExist:
        return render_to_response('status.html', {'submission':submission})

    # If the submission was processed successfully...
    if job.deleted:
        return render_to_response('deleted.html')
    if job.ok:
        downloadsUrl = ''.join(["/", settings.ZIP_DOWNLOAD_URL,
                               str(job.submission.url_id),
                               ".zip"])
        imageUrl = ''.join(["/", settings.PREVIEW_URL,
                             str(job.submission.url_id),
                            "/tiles/0_0_0.png"])
        deleteUrl = ''.join(["/delete/", job.submission.user_admin_id])
        return render_to_response('finished.html',
                                  {'downloadsUrl' : downloadsUrl,
                                   'previewUrl' : submission.get_absolute_url(),
                                   'imageUrl' : imageUrl,
                                   'deleteUrl' : deleteUrl})
    else:
        return render_to_response('failed_submission.html',
                                  {'email': settings.CONTACT_EMAIL})
